/*******************************************************************************
***  This file is part of Qt Mocks.                                          ***
***                                                                          ***
***  Copyright (C) 2018, 2019, 2020, 2021, 2023, 2024                        ***
***  John Schneiderman <Licencing _AT_ Schneiderman _DOT_ me>                ***
***                                                                          ***
***  This program is free software: you can redistribute it and/or modify it ***
***  under the terms of the GNU Lesser General Public License as published   ***
***  by the Free Software Foundation, either version 3 of the License, or    ***
***  (at your option) any later version.                                     ***
***                                                                          ***
***  This program is distributed in the hope that it will be useful, but     ***
***  WITHOUT ANY WARRANTY; without even the implied warranty of              ***
***  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    ***
***  See the GNU Lesser General Public License for more details.             ***
***                                                                          ***
***  You should have received a copy of the GNU Lesser General Public License***
***  along with this program. If not, see <http://www.gnu.org/licenses/>.    ***
*******************************************************************************/
#include "Expectation.hpp"
using ::QTest::qFail;
using qt_mocks::internals::Expectation;
using qt_mocks::internals::ExpectationBuilder;
using qt_mocks::internals::ExpectationQueue;
using qt_mocks::internals::ExpectedAction;
using qt_mocks::internals::toStdString;
using ::quint64;
using std::move;
using std::nullopt;
using std::optional;
using std::unique_ptr;

#include <QString>
using ::QString;
// using QStringLiteral

#include <exception>
using std::exception;
#include <stdexcept>
using std::invalid_argument;

//{ Expectation

Expectation::Expectation(
	const char* const functionName,
	const char* const fileName,
	const int lineNumber
) :
	functionName_{functionName},
	fileName_{fileName},
	lineNumber_{lineNumber},
	isRequired_{false},
	expectedInvokeCount_{nullopt},
	actualInvokedCount_{0u},
	action_{ExpectedAction::Invoked}
{}

Expectation::~Expectation() = default;

const char* Expectation::fileName() const noexcept
{
	return this->fileName_;
}

int Expectation::lineNumber() const noexcept
{
	return this->lineNumber_;
}

const char* Expectation::functionName() const noexcept
{
	return this->functionName_;
}

bool Expectation::isInvokeRequired() const noexcept
{
	return this->isRequired_;
}

const optional<quint64>& Expectation::expectedInvokeCount() const noexcept
{
	return this->expectedInvokeCount_;
}

quint64 Expectation::actualInvokedCount() const noexcept
{
	return this->actualInvokedCount_;
}

QString Expectation::description() const noexcept
// cppcheck-suppress syntaxError
// Justification: It is in fact, not a syntax error.
try
{
	if (this->expectedInvokeCount_.has_value())
		return QStringLiteral(
			"expected '%1' to %2. Expected %3 invocations, but was invoked %4 times."
		).arg(
			this->functionName(),
			toStdString(this->action_),
			QString::number(*this->expectedInvokeCount_),
			QString::number(this->actualInvokedCount_)
		);
	return QStringLiteral("expected '%1' to %2; it was invoked %3 times.").arg(
		this->functionName(),
		toStdString(this->action_),
		QString::number(this->actualInvokedCount_)
	);
}
catch (const exception& error)
{
	return QStringLiteral(
		"expected '%1' to perform an action; it was invoked %2 times. Unexpected error: %3."
	).arg(
		this->functionName(),
		QString::number(this->actualInvokedCount_),
		error.what()
	);
}

void Expectation::markInvokeRequired() noexcept
{
	this->isRequired_ = true;
}

void Expectation::setExpectedInvokeCount(const quint64 count) noexcept
{
	this->expectedInvokeCount_.emplace(count);
}

void Expectation::increaseInvokeCount() noexcept
{
	++this->actualInvokedCount_;
}

//}

//{ QueuedExpectations

#ifndef QTEST_FAIL_ACTION
#	define QTEST_FAIL_ACTION do{ ; }while(false)
#endif

ExpectationQueue::~ExpectationQueue()
{
	bool hasUnmetRequiredExpectations{false};

	while (! this->expectations_.empty())
	{
		if (
			const auto& expectation{this->expectations_.front()} ;
			expectation->actualInvokedCount() == 0
			|| (
				expectation->expectedInvokeCount().has_value()
				&& expectation->actualInvokedCount() != *expectation->expectedInvokeCount()
			)
		)
		{
			if (expectation->isInvokeRequired())
			{
				qFail(
					QStringLiteral("Unmet expectation: ").append(expectation->description()).
					toUtf8().constData(),
					expectation->fileName(),
					expectation->lineNumber()
				);
				hasUnmetRequiredExpectations = true;
			}
			else
				QMessageLogger(
					expectation->fileName(),
					expectation->lineNumber(),
					expectation->functionName()
				).warning() << QStringLiteral("Unmet expectation: ").append(
					expectation->description()
				).toUtf8().constData();
		}
		this->expectations_.pop();
	}

	if (hasUnmetRequiredExpectations)
		QTEST_FAIL_ACTION;
}

const unique_ptr<Expectation>& ExpectationQueue::peek() const noexcept
{
	static const unique_ptr<Expectation> empty{nullptr};

	if (this->expectations_.empty())
		return empty;
	return this->expectations_.front();
}

Expectation& ExpectationQueue::push(unique_ptr<Expectation> expectation)
{
	this->expectations_.push(move(expectation));
	return *this->expectations_.back();
}

void ExpectationQueue::pop()
{
	this->expectations_.pop();
}

//}

//{ Builder Without Arguments & No Return

ExpectationBuilder<void, void>::ExpectationBuilder(Expectation& expectation) :
	expectation_{dynamic_cast<MockedExpectation<void, void>&>(expectation)}
{}

ExpectationBuilder<void, void>& ExpectationBuilder<void, void>::times(const quint64 count)
{
	this->expectation_.setExpectedInvokeCount(count);
	return *this;
}

ExpectationBuilder<void, void>& ExpectationBuilder<void, void>::require()
{
	this->expectation_.markInvokeRequired();
	return *this;
}

//}

QString qt_mocks::internals::toStdString(const ExpectedAction action)
{
	switch (action)
	{
	case ExpectedAction::Invoked:
		return QStringLiteral("be invoked");
	case ExpectedAction::GiveValue:
		return QStringLiteral("be invoked, and give a value");
	case ExpectedAction::RaiseError:
		return QStringLiteral("be invoked, and raise an error");
	case ExpectedAction::GenerateValue:
		return QStringLiteral("be invoked, and give a generated value");
	default:
		throw invalid_argument{"Invalid action supplied."};
	}
}
